from collections import deque
from functools import partialmethod
import math
import re

from regularize.exceptions import SampleNotMatchedError, \
    InvalidRangeError
from regularize.flag import FlagSet


class Metacharacter:
    def __copy__(self):
        return self.__class__()

    def __repr__(self):
        return f'\'{self.__class__.__name__} -> {str(self)}\''


class OpeningBracket(Metacharacter):
    def __str__(self):
        return '['


class ClosingBracket(Metacharacter):
    def __str__(self):
        return ']'


class Expression:
    def __init__(self, parent: 'Expression' = None):
        self._token_stack = deque()
        self._bracket_stack = []
        if parent:
            self._copy_state(parent)

    def _copy_state(self, other, clear=True):
        if clear:
            self.bracket_stack.clear()
            self.token_stack.clear()
        self.bracket_stack.extend(other.bracket_stack)
        self.token_stack.extend(other.token_stack)

    @property
    def token_stack(self) -> deque:
        return self._token_stack

    @property
    def bracket_stack(self) -> list:
        return self._bracket_stack

    def has_open_bracket(self):
        if not self.bracket_stack:
            return False
        return isinstance(self.bracket_stack[-1], OpeningBracket)

    def close_bracket(self):
        if not self.has_open_bracket():
            return self

        if self.bracket_stack:
            last_item_in_stack = self.bracket_stack[-1]
        else:
            last_item_in_stack = None

        if not isinstance(last_item_in_stack, OpeningBracket):
            raise RuntimeError('Cannot close bracket without opening')

        return self.clone_with_updates(append=ClosingBracket())

    def _prepare_for_build(self):
        return self.close_bracket()

    def build(self):
        return ''.join(map(str, self._prepare_for_build().token_stack))

    def __repr__(self):
        return f"{self.__class__.__name__}<{hex(id(self))}>[{self.token_stack}]"

    def __str__(self):
        return f'Expression: /{self.build()}/'

    def __add__(self, other):
        new = self.__class__(parent=self)
        new._copy_state(other, clear=False)
        return new

    def clone(self) -> 'Expression':
        new = self.__class__(parent=self)
        self._on_after_clone(new)
        return new

    def _on_after_clone(self, new):
        pass

    def clone_with_updates(self, append=None, prepend=None) -> 'Expression':
        if append is not None and not isinstance(append, (list, tuple)):
            append = (append,)

        if prepend is not None and not isinstance(prepend, (list, tuple)):
            prepend = (prepend,)

        clone = self.clone()
        clone.token_stack.extendleft(reversed(prepend or []))
        clone.token_stack.extend(append or [])

        if append:
            for item in append:
                if isinstance(item, ClosingBracket):
                    if clone.has_open_bracket():
                        clone.bracket_stack.pop()
                elif isinstance(item, OpeningBracket):
                    clone.bracket_stack.append(item)

        return clone


class Pattern(Expression):
    def __init__(self, *args, **kwargs):
        self._extensions = None
        self._flags = None
        super(Pattern, self).__init__(*args, **kwargs)

    def _copy_state(self, other, clear=True):
        super(Pattern, self)._copy_state(other, clear=clear)
        self._extensions = other.extensions.clone()

    def __eq__(self, other):
        return self.flags.equals(other.flags) and \
               self.token_stack == other.token_stack

    @property
    def flags(self):
        if self._flags is None:
            self._flags = FlagSet(pattern=self)
        return self._flags

    def _on_after_clone(self, new):
        new._flags = FlagSet.copy(pattern=new)
        new._extensions = self.extensions.clone()

    def group(self, name=None, optional=False, pattern=None):
        if pattern is None:
            wrapped_pattern = self
        else:
            wrapped_pattern = pattern

        if name is None:
            new_group = Group(wrapped_pattern)(optional=optional)
        else:
            new_group = NamedGroup(wrapped_pattern)(name, optional=optional)

        if pattern is None:
            return new_group
        else:
            return self + new_group

    def match_any(self, *subexpressions, **kwargs):
        expression_list = [
            subexpression.build()
            for subexpression in map(self._ensure_pattern, subexpressions)
        ]
        new = self.__class__()
        new = new.raw('|'.join(expression_list)).group(**kwargs)
        return self.clone_with_updates(new.build())

    @staticmethod
    def _ensure_pattern(obj):
        if isinstance(obj, str):
            return Literal()(obj)
        elif isinstance(obj, Pattern):
            return obj
        else:
            raise TypeError(f'Cannot handle type {obj.__class__.__name__} automatically')

    def __or__(self, other):
        return (self.clone_with_updates(append='|') + other).group()

    def whitespace(self, match=True) -> 'Pattern':
        return Whitespace(self)(match)

    def lowercase_ascii_letters(self, **kwargs):
        return AsciiLetterCharacter(self)(lowercase=True, **kwargs)
    # Alias due to high frequency use (along with case-insensitive flag)
    ascii_letters = lowercase_ascii_letters

    def uppercase_ascii_letters(self, **kwargs) -> 'Pattern':
        return AsciiLetterCharacter(self)(lowercase=False, **kwargs)

    def any_number_between(self, **kwargs):
        return Number(self)(**kwargs)
    any_number = any_number_between

    def quantify(self, minimum=0, maximum=math.inf):
        addition = None
        if minimum == 0 and math.isinf(maximum):
            addition = '*'
        elif minimum == 0 and maximum == 1:
            addition = '?'
        elif minimum == 1 and math.isinf(maximum):
            addition = '+'
        elif minimum == maximum:
            addition = f'{{{minimum}}}'
        elif minimum > 1 and math.isinf(maximum):
            addition = f'{{{minimum},}}'
        elif not math.isinf(maximum):
            addition = f'{{{minimum},{maximum}}}'
        return self.close_bracket().clone_with_updates(append=addition)
    at_least_one = partialmethod(quantify, minimum=1, maximum=math.inf)

    def exactly(self, times):
        return self.quantify(minimum=times, maximum=times)

    def wildcard(self, one_or_more=False):
        if one_or_more:
            add = '.+'
        else:
            add = '.'
        return self.clone_with_updates(add)
    match_all = partialmethod(wildcard, one_or_more=True)

    def literal(self, string):
        return Literal(self)(string)

    def any_of(self, *members, close=True):
        clone = self.clone_with_updates(append=OpeningBracket())
        if members:
            expression = ''.join(
                map(
                    str,
                    map(BracketExpressionPartial.ensure, members)
                )
            )
            clone = clone.clone_with_updates(expression)
        if close:
            clone = clone.close_bracket()
        return clone

    def none_of(self, *members):
        clone = self.clone_with_updates(append=OpeningBracket())
        if members:
            expression = ''.join(
                map(
                    str,
                    map(BracketExpressionPartial.ensure, members)
                )
            )
            clone = clone.clone_with_updates(
                f"^{expression}"
            )
        return clone

    def raw(self, string):
        return self.clone_with_updates(string)

    def start_anchor(self):
        return self.clone_with_updates(append='^')

    def end_anchor(self):
        return self.clone_with_updates(append='$')

    def compile(self):
        try:
            return re.compile(self.build(), self.flags.compile())
        except re.error as e:
            print(f'Unable to build regular expression: {self}')
            raise e

    def test(self, sample):
        regex = self.compile()
        match = regex.match(sample)
        if not match:
            raise SampleNotMatchedError(f'{regex} tested with "{sample}"')
        return match

    def case_insensitive(self, enabled=True):
        return self.clone().flags.case_insensitive(enabled=enabled)

    def multiline(self, enabled=True):
        return self.clone().flags.multiline(enabled=enabled)

    def dot_matches_newline(self, enabled=True):
        return self.clone().flags.dot_matches_newline(enabled=enabled)

    def ascii_only(self, enabled=True):
        return self.clone().flags.ascii_only(enabled=enabled)

    def __str__(self):
        initial = super(Pattern, self).__str__()
        return f'{initial}{self.flags}'

    @property
    def ext(self):
        if self._extensions is None:
            self._extensions = Extensions(self)
        return self._extensions
    extensions = ext

    @classmethod
    def join(cls, delimiter, subpatterns):
        composite_pattern = cls()
        for subpattern in subpatterns[:-2]:
            composite_pattern = composite_pattern + subpattern + delimiter
        composite_pattern = composite_pattern + subpatterns[-1]
        return composite_pattern


class Group(Pattern):
    def __call__(self, optional) -> 'Pattern':
        append_additions = [')']
        if optional:
            append_additions.append('?')
        return self.close_bracket().clone_with_updates(
            prepend='(',
            append=append_additions
        )


class NamedGroup(Group):
    def __call__(self, name, optional) -> 'Pattern':
        append_additions = [')']
        if optional:
            append_additions.append('?')
        return self.close_bracket().clone_with_updates(
            prepend=('(', f'?P<{name}>'),
            append=append_additions
        )


class Literal(Pattern):
    def __call__(self, string):
        return self.clone_with_updates(re.escape(string))


class Whitespace(Pattern):
    def __call__(self, match):
        return self.clone_with_updates('\\s' if match else '\\S')


class Range(Pattern):
    def __call__(self, start, end, closed=False, negated=False, skip_brackets=False):
        if negated:
            start = f'^{start}'

        additions = []
        if not self.has_open_bracket() and not skip_brackets:
            additions.append(OpeningBracket())
        additions.append(f'{start}-{end}')
        if closed and not skip_brackets:
            additions.append(ClosingBracket())
        return self.clone_with_updates(append=additions)


class AsciiLetterCharacter(Range):
    def __call__(self, lowercase=True, **kwargs):
        start = 'a' if lowercase else 'A'
        end = 'z' if lowercase else 'Z'
        return super(AsciiLetterCharacter, self).__call__(start, end, **kwargs)


class Number(Range):
    def __call__(self, minimum=0, maximum=9, **kwargs) -> Pattern:
        if minimum >= maximum or minimum < 0 or maximum > 9:
            raise InvalidRangeError(
                f'Cannot build range between {minimum} and {maximum}'
            )
        return super(Number, self).__call__(minimum, maximum, **kwargs)


class BracketExpressionPartial:
    def __init__(self, expression: str):
        self._expression = expression

    def __str__(self):
        return self._expression

    def __repr__(self):
        return f'{self.__class__.__name__}: {repr(self._expression)}'

    @property
    def expression(self):
        return self._expression

    @classmethod
    def ensure(cls, obj):
        if not isinstance(obj, cls):
            return cls(Literal()(obj).build())
        else:
            return obj


class ExtensionRegistry:
    def __init__(self):
        self._registrations = {}

    @property
    def registrations(self):
        return self._registrations

    def add(self, name, klass):
        self._registrations[name] = klass

    def __iter__(self):
        for registration in self._registrations.items():
            yield registration

    def __contains__(self, item):
        return item in self._registrations

    def __repr__(self):
        return repr(self._registrations)

    def clear(self):
        self._registrations.clear()


Pattern.registry = ExtensionRegistry()


class Extensions:
    _registry = Pattern.registry

    def __init__(self, pattern):
        self._pattern = pattern
        self._callbacks_initialized = False
        self._callbacks = {}

    @property
    def registry(self):
        return self.__class__._registry

    def clone(self):
        new = self.__class__(self._pattern)
        new._callbacks_initialized = False
        return new

    def _initialize_callbacks(self):
        if self._callbacks_initialized:
            return

        for name, klass in self.registry:
            self._callbacks[name] = klass(self._pattern)
        self._callbacks_initialized = True

    def __getattr__(self, item):
        if item in self.registry:
            self._initialize_callbacks()
            return self._callbacks[item]
        else:
            raise AttributeError(item)


Pattern.ANY_NUMBER = BracketExpressionPartial(Number()(skip_brackets=True).build())
Pattern.ANY_ASCII_CHARACTER = BracketExpressionPartial(
    AsciiLetterCharacter()(skip_brackets=True).build()
)
Pattern.NO_WHITESPACE = BracketExpressionPartial(Whitespace()(match=False).build())
Pattern.ANY_WHITESPACE = BracketExpressionPartial(Whitespace()(match=True).build())

pattern = Pattern



from functools import reduce
from operator import or_
import re
import typing

if typing.TYPE_CHECKING:
    from regularize.expression import Pattern


class FlagSet:
    def __init__(self, pattern: 'Pattern' = None):
        self._options = set()
        self._pattern = pattern

    @classmethod
    def copy(cls, pattern: 'Pattern' = None):
        new = cls(pattern=pattern)
        if pattern is not None:
            new._options.update(pattern.flags.options)
        return new

    @property
    def options(self):
        return self._options

    @property
    def pattern(self):
        return self._pattern

    def _add_option(self, flag):
        self.options.add(flag)
        return self.pattern or self

    def _remove_option(self, flag):
        self.options.remove(flag)
        return self.pattern or self

    def _update_option(self, flag, enabled):
        if enabled:
            return self._add_option(flag)
        else:
            return self._remove_option(flag)

    def case_insensitive(self, enabled=True):
        return self._update_option(re.IGNORECASE, enabled=enabled)

    def ascii_only(self, enabled=True):
        return self._update_option(re.ASCII, enabled=enabled)

    def multiline(self, enabled=True):
        return self._update_option(re.MULTILINE, enabled=enabled)

    def dot_matches_newline(self, enabled=True):
        return self._update_option(re.DOTALL, enabled=enabled)

    def compile(self):
        if self._options:
            return reduce(or_, self._options, 0)
        return 0

    def __eq__(self, other):
        if self.pattern is None:
            pattern_equal = other.pattern is None
        else:
            pattern_equal = self.pattern == other.pattern
        return pattern_equal and self.equals(other)

    def equals(self, other):
        return self.options == other.options

    def __str__(self):
        if self._options:
            return repr(self.options)
        else:
            return ''

    def __repr__(self):
        return f'{self.__class__.__name__}: {repr(self.options)}'



from functools import wraps
from typing import Union
import re
import typing


if typing.TYPE_CHECKING:
    from regularize.expression import Pattern


class Cache:
    NOT_FOUND = object()
    DEFAULT_MAXIMUM_SIZE = 1_000

    def __init__(self, maxsize=DEFAULT_MAXIMUM_SIZE):
        self._cache = dict()
        self._maxsize = maxsize
        self._reset_stats()

    def _increment_metric(self, name):
        self._stats[name] += 1

    def _reset_stats(self):
        self._stats = {
            'maxsize_reached': 0,
            'hits': 0,
            'misses': 0
        }

    @property
    def stats(self):
        return self._stats.copy()

    @property
    def cache(self):
        return self._cache.copy()

    @property
    def current_size(self):
        return len(self._cache)

    def clear(self):
        self._reset_stats()
        self._cache.clear()

    def get(self, key):
        if key in self._cache:
            entry = self._cache[key]
        else:
            entry = self.NOT_FOUND
            self._increment_metric('misses')

        if entry is not self.NOT_FOUND:
            self._increment_metric('hits')
            # Emulate LRU by utilizing LIFO order in dictionaries.
            # Note that this has performance impact, but does not
            # require maintaining extra statistics or structures.
            del self._cache[key]
            self.add(key, entry)

        return entry, entry is not self.NOT_FOUND

    def add(self, key, entry):
        self._cache.setdefault(key, entry)
        if len(self._cache) > self._maxsize:
            self._increment_metric('maxsize_reached')
            # Remove the first key which should be the least recently
            # accessed. See .get for details.
            remove_key = next(iter(self._cache.keys()))
            del self._cache[remove_key]


def enable_dict_cache(maxsize):
    cache = Cache(maxsize=maxsize)

    def cached(func):
        @wraps(func)
        def cached_wrapper(cls, pattern, string):
            key = (pattern, string)
            entry, found = cache.get(key)
            if found:
                return entry
            entry = func(cls, pattern, string)
            cache.add(key, entry)
            return entry
        cached_wrapper._cache = cache
        return cached_wrapper
    return cached


class Finder:
    def __init__(self, pattern: Union['Pattern', re.Pattern]):
        self._pattern = pattern
        self._compiled_pattern = None
        self._is_builtin_pattern = isinstance(pattern, re.Pattern)

    @property
    def pattern(self):
        return self._pattern

    @property
    def compiled_pattern(self):
        if self._is_builtin_pattern:
            return self.pattern

        if self._compiled_pattern is None:
            self._compiled_pattern = self.pattern.compile()

        return self._compiled_pattern

    def match(self, string):
        return self.__class__._match(self.compiled_pattern, string)

    def find(self, string, iterator=True):
        if iterator:
            return self.compiled_pattern.finditer(string)
        else:
            return self.compiled_pattern.findall(string)

    @classmethod
    @enable_dict_cache(maxsize=1_000)
    def _match(cls, regex, string):
        return regex.match(string)

    @classmethod
    def cache_clear(cls):
        # TODO: avoid access to protected methods
        return cls._match._cache.clear()


finder = Finder



from functools import lru_cache
import typing

if typing.TYPE_CHECKING:
    from regularize.expression import Pattern


class Substitution:
    def __init__(self, pattern: 'Pattern'):
        self._stack = []
        self._pattern = pattern
        self._compiled_pattern = None

    @property
    def pattern(self):
        if self._compiled_pattern is None:
            self._compiled_pattern = self._pattern.compile()
        return self._compiled_pattern

    @property
    def stack(self):
        return self._stack

    def _build(self):
        return ''.join(self.stack)

    def add(self, string):
        self.stack.append(string)
        return self

    def backreference(self, name_or_number):
        self.stack.append(f'\\g<{name_or_number}>')
        return self

    @lru_cache(maxsize=1_000)
    def replace(self, string, count=0):
        return self.pattern.sub(self._build(), string, count=count)
    __call__ = replace


substitution = Substitution



from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name="regularize",
    version="0.0.5",
    author="George Psarakis",
    description="Regular Expression Builder",
    long_description=readme,
    long_description_content_type="text/markdown",
    package_dir={"": "."},
    packages=find_packages(),
    zip_safe=False,
    install_requires=[],
    extras_require={
        "dev": [
            "pytest==6.1.0",
            "flake8==3.8.4",
            "pytest-cov==2.10.1"
        ]
    },
    url="https://github.com/georgepsarakis/regularize",
    license="MIT",
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)
